/* A-626 Service Module.

   Author: Reed Lai

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. */

/* File level history (record changes for this file here.)

   v 0.5.0 3  Sep 2013 by Winson Chen
     Give support to linux kernel 3.X.X.

   v 0.4.0 18 Aug 2011 by Golden Wang
     Give support to linux kernel 2.6.X.

   v 0.3.0 18 Nov 2002 by Reed Lai
     Comments out the including of slab.h which is no longer used
     by this module.

     Uses slab.h in place of malloc.h.
     
   v 0.2.0 27 Feb 2002 by Reed Lai
     Use a new flag type for device information (ixisa_devinfo.)

   v 0.1.0 11 Jan 2002 by Reed Lai
     Rename DAQ to IXISA.

   v 0.0.0 24 Jul 2001 by Reed Lai
     create, blah blah... */

/* Mandatory */
#include <linux/kernel.h>		/* ta, kernel work */
#include <linux/module.h>		/* is a module */
#include "ixisa.h"

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
        /* Deal with CONFIG_MODVERSIONS that is defined in
         * /usr/include/linux/config.h (config.h is included by module.h) */
        #if CONFIG_MODVERSIONS==1
        #define MODVERSIONS
        #include <linux/modversions.h>
        #endif

        /* for compatibility with future version of Linux */
        #include <linux/wrapper.h>
#endif

/* Additional */
#include <linux/fs.h>

/* use I/O ports */
#include <asm/io.h>
#include <linux/ioport.h>

/* need kmalloc */
#include <linux/slab.h>

/* irq service */
#include <linux/sched.h>
#include <linux/interrupt.h>
#include <asm/irq.h>

/* Local */
#include "_flag_inline.h"
#include "_a626.h"

#define BUF_LEN 80

char msg[BUF_LEN];
char *msg_ptr;

ixisa_devinfo_t *dev;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,36)
int ixisa_minor[256] = {0};
#endif

#ifdef MODULE_LICENSE
MODULE_AUTHOR("Golden Wang <services@icpdas.com>");
MODULE_DESCRIPTION("ICPDAS ISA-series driver, A-626 service module");
MODULE_LICENSE(ICPDAS_LICENSE);
#endif  /* MODULE_LICENSE */


/* symbols exportation */
// EXPORT_NO_SYMBOLS; /* do not export symbols */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
EXPORT_SYMBOL_GPL(ixa626_ioctl);
EXPORT_SYMBOL_GPL(ixa626_release);
EXPORT_SYMBOL_GPL(ixa626_open);
#else
EXPORT_SYMBOL_NOVERS(ixa626_ioctl);
EXPORT_SYMBOL_NOVERS(ixa626_release);
EXPORT_SYMBOL_NOVERS(ixa626_open);
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,31)
struct task_struct *find_task_by_pid(pid_t nr, struct pid_namespace *ns)
{
        return pid_task(find_pid_ns(nr, ns), PIDTYPE_PID);
}
#endif

#ifdef BOTTOM_HELF
void irq_handler_bh(void *data)
{
	/* bottom helf of irq handler
	 *
	 * Arguments:
	 *   data     data pointer as <linux/tqueue.h> defined
	 *
	 * Returned:  none */

	ixisa_devinfo_t *dp;

	dp = (ixisa_devinfo_t *) data;

	/* reserved */
}
#endif							/* BOTTOM_HELF */

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,22)
  irqreturn_t irq_handler(int irq, void *dev_id)
#else
  #if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
  irqreturn_t irq_handler(int irq, void *dev_id, struct pt_regs *regs)
  #else
  void irq_handler(int irq, void *dev_id, struct pt_regs *regs)
  #endif
#endif
{
	/* Arguments:  as <linux/sched.h> defined
	 *
	 * Returned:  none */

	ixisa_devinfo_t *dp;

#ifdef BOTTOM_HELF
	static struct tq_struct task = { NULL, 0, irq_handler_bh, NULL };

	task.data = dev_id;
#endif							/* BOTTOM_HELF */

	dp = (ixisa_devinfo_t *) dev_id;


        if (dp->sig.sid)
        {
                 send_sig(dp->sig.sid, dp->sig.task, 1);
        }

	/* Scheduale bottom half to run */
#ifdef BOTTOM_HELF
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,2,0)
	queue_task(&task, &tq_immediate);
#else
	queue_task_irq(&task, &tq_immediate);
#endif
	mark_bh(IMMEDIATE_BH);
#endif							/* BOTTOM_HELF */

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
        return IRQ_HANDLED;
#endif
}

int set_signal(ixisa_signal_t * sig, ixisa_devinfo_t * dp)
{
	/* set user signal for interrupt
	 *
	 * Arguments:
	 *   sig      pointer to structure ixisa_signal for user signal
	 *   di       pointer to structure ixisa_devinfo for current device
	 *
	 * Returned:  SUCCESS or FAILURE */

	if (sig->sid) {
		/* has signal id, set */
		if (sig->pid) {
			/* user gives his process id */
			dp->sig.sid = sig->sid;
			dp->sig.pid = sig->pid;

			#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,31)
                        dp->sig.task = find_task_by_pid(sig->pid,&init_pid_ns);

                        #elif LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,26)
                        dp->sig.task = find_task_by_pid_type_ns(PIDTYPE_PID,sig->pid,&init_pid_ns);
                        #else
                        dp->sig.task = find_task_by_pid(sig->pid);
                        #endif
		} else if (sig->task) {
			/* user gives his task structure */
			dp->sig.sid = sig->sid;
			dp->sig.task = sig->task;
		} else {
			/* wrong! */
			return FAILURE;
		}
		dp->sig.is = sig->is;
		dp->sig.edge = sig->edge;
		dp->sig.bedge = sig->bedge;
	} else {
		/* no signal id, clear */
		dp->sig.sid = 0;
		dp->sig.pid = 0;
		dp->sig.task = 0;
		dp->sig.is = 0;
		dp->sig.edge = 0;
		dp->sig.bedge = 0;
	}
	return SUCCESS;
}

int write_reg(ixisa_reg_t * reg, unsigned long int base)
{
	/* Write to register
	 *
	 * Arguments:
	 *   reg      pointer to a structure ixisa_reg for register
	 *   base     base address for current device
	 *
	 * Returned:  SUCCESS or FAILURE */

	unsigned char lbyte, hbyte;

	switch (reg->id) {
	case IXISA_AO0H:
#if AO0H_MASK == 0xff
		outb((unsigned char) reg->value, base + AO0H);
#else
		outb((unsigned char) reg->value & AO0H_MASK, base + AO0H);
#endif
		break;
	case IXISA_AO0L:
#if AO0L_MASK == 0xff
		outb((unsigned char) reg->value, base + AO0L);
#else
		outb((unsigned char) reg->value & AO0L_MASK, base + AO0L);
#endif
		break;
	case IXISA_AO0:
		/* Analog Output (D/A) Latch Register */
		lbyte = (unsigned char) reg->value;
		hbyte = (unsigned char) (reg->value >> 8);
#if AO0H_MASK == 0xff
		outb(hbyte, base + AO0H);
#else
		outb(hbyte & AO0H_MASK, base + AO0H);
#endif
#if AO0L_MASK == 0xff
		outb(lbyte, base + AO0L);
#else
		outb(lbyte & AO0L_MASK, base + AO0L);
#endif
		break;
	case IXISA_AO1H:
#if AO1H_MASK == 0xff
		outb((unsigned char) reg->value, base + AO1H);
#else
		outb((unsigned char) reg->value & AO1H_MASK, base + AO1H);
#endif
		break;
	case IXISA_AO1L:
#if AO1L_MASK == 0xff
		outb((unsigned char) reg->value, base + AO1L);
#else
		outb((unsigned char) reg->value & AO1L_MASK, base + AO1L);
#endif
		break;
	case IXISA_AO1:
		/* Analog Output (D/A) Latch Register */
		lbyte = (unsigned char) reg->value;
		hbyte = (unsigned char) (reg->value >> 8);
#if AO1H_MASK == 0xff
		outb(hbyte, base + AO1H);
#else
		outb(hbyte & AO1H_MASK, base + AO1H);
#endif
#if AO1L_MASK == 0xff
		outb(lbyte, base + AO1L);
#else
		outb(lbyte & AO1L_MASK, base + AO1L);
#endif
		break;
	case IXISA_AO2H:
#if AO2H_MASK == 0xff
		outb((unsigned char) reg->value, base + AO2H);
#else
		outb((unsigned char) reg->value & AO2H_MASK, base + AO2H);
#endif
		break;
	case IXISA_AO2L:
#if AO2L_MASK == 0xff
		outb((unsigned char) reg->value, base + AO2L);
#else
		outb((unsigned char) reg->value & AO2L_MASK, base + AO2L);
#endif
		break;
	case IXISA_AO2:
		/* Analog Output (D/A) Latch Register */
		lbyte = (unsigned char) reg->value;
		hbyte = (unsigned char) (reg->value >> 8);
#if AO2H_MASK == 0xff
		outb(hbyte, base + AO2H);
#else
		outb(hbyte & AO2H_MASK, base + AO2H);
#endif
#if AO2L_MASK == 0xff
		outb(lbyte, base + AO2L);
#else
		outb(lbyte & AO2L_MASK, base + AO2L);
#endif
		break;
	case IXISA_AO3H:
#if AO3H_MASK == 0xff
		outb((unsigned char) reg->value, base + AO3H);
#else
		outb((unsigned char) reg->value & AO3H_MASK, base + AO3H);
#endif
		break;
	case IXISA_AO3L:
#if AO3L_MASK == 0xff
		outb((unsigned char) reg->value, base + AO3L);
#else
		outb((unsigned char) reg->value & AO3L_MASK, base + AO3L);
#endif
		break;
	case IXISA_AO3:
		/* Analog Output (D/A) Latch Register */
		lbyte = (unsigned char) reg->value;
		hbyte = (unsigned char) (reg->value >> 8);
#if AO3H_MASK == 0xff
		outb(hbyte, base + AO3H);
#else
		outb(hbyte & AO3H_MASK, base + AO3H);
#endif
#if AO3L_MASK == 0xff
		outb(lbyte, base + AO3L);
#else
		outb(lbyte & AO3L_MASK, base + AO3L);
#endif
		break;
	case IXISA_AO4H:
#if AO4H_MASK == 0xff
		outb((unsigned char) reg->value, base + AO4H);
#else
		outb((unsigned char) reg->value & AO4H_MASK, base + AO4H);
#endif
		break;
	case IXISA_AO4L:
#if AO4L_MASK == 0xff
		outb((unsigned char) reg->value, base + AO4L);
#else
		outb((unsigned char) reg->value & AO4L_MASK, base + AO4L);
#endif
		break;
	case IXISA_AO4:
		/* Analog Output (D/A) Latch Register */
		lbyte = (unsigned char) reg->value;
		hbyte = (unsigned char) (reg->value >> 8);
#if AO4H_MASK == 0xff
		outb(hbyte, base + AO4H);
#else
		outb(hbyte & AO4H_MASK, base + AO4H);
#endif
#if AO4L_MASK == 0xff
		outb(lbyte, base + AO4L);
#else
		outb(lbyte & AO4L_MASK, base + AO4L);
#endif
		break;
	case IXISA_AO5H:
#if AO5H_MASK == 0xff
		outb((unsigned char) reg->value, base + AO5H);
#else
		outb((unsigned char) reg->value & AO5H_MASK, base + AO5H);
#endif
		break;
	case IXISA_AO5L:
#if AO5L_MASK == 0xff
		outb((unsigned char) reg->value, base + AO5L);
#else
		outb((unsigned char) reg->value & AO5L_MASK, base + AO5L);
#endif
		break;
	case IXISA_AO5:
		/* Analog Output (D/A) Latch Register */
		lbyte = (unsigned char) reg->value;
		hbyte = (unsigned char) (reg->value >> 8);
#if AO5H_MASK == 0xff
		outb(hbyte, base + AO5H);
#else
		outb(hbyte & AO5H_MASK, base + AO5H);
#endif
#if AO5L_MASK == 0xff
		outb(lbyte, base + AO5L);
#else
		outb(lbyte & AO5L_MASK, base + AO5L);
#endif
		break;
	case IXISA_DO_A:			/* DO 0-7 */
#if DO_A_MASK == 0xff
		outb(((unsigned char) reg->value), base + DO_A);
#else
		outb(((unsigned char) reg->value) & DO_A_MASK, base + DO_A);
#endif
		break;
	case IXISA_DO_B:			/* DO 8-15 */
#if DO_B_MASK == 0xff
		outb(((unsigned char) reg->value), base + DO_B);
#else
		outb(((unsigned char) reg->value) & DO_B_MASK, base + DO_B);
#endif
		break;
	default:
		return FAILURE;
	}
	return SUCCESS;
}

int read_reg(ixisa_reg_t * reg, unsigned long int base)
{
	/* Read from register
	 *
	 * Arguments:
	 *   reg      pointer to structure ixisa_reg for register
	 *   base     base address for current device
	 *
	 * Returned:  SUCCESS or FAILURE */

	switch (reg->id) {
	case IXISA_DI_A:			/* DI 0-7 */
#if DI_A_MASK == 0xff
		reg->value = inb(base + DI_A);
#else
		reg->value = inb(base + DI_A) & DI_A_MASK;
#endif
		break;
	case IXISA_DI_B:			/* DI 8-15 */
#if DI_B_MASK == 0xff
		reg->value = inb(base + DI_B);
#else
		reg->value = inb(base + DI_B) & DI_B_MASK;
#endif
		break;
	default:
		return FAILURE;
	}
	return SUCCESS;
}

ixisa_devinfo_t *align_to_minor(int minor)
{
	ixisa_devinfo_t *p;
	for (p = dev; p && p->no != minor && p->next_f; p = p->next_f);
	if (p->no == minor)
		return p;
	return 0;
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,36)
long ixa626_ioctl(struct file *file, unsigned int ioctl_num, unsigned long ioctl_param)
#else
int ixa626_ioctl(struct inode *inode, struct file *file, unsigned int ioctl_num, unsigned long ioctl_param)
#endif
{
	/* (export)
	 *
	 * This function is called by ixisa.o whenever a process tries to do and
	 * IO control on DIO device file
	 *
	 * Arguments: read <linux/fs.h> for (*ioctl) of struct file_operations
	 *
	 * Returned:  SUCCESS or FAILED */

	ixisa_devinfo_t *dp;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,36)
        int minor;

        if(file->private_data)
        {
                minor = *((int *)file->private_data);
        }
        else
        {
                return -EINVAL;
        }

        dp = align_to_minor(minor);
#else
	dp = align_to_minor(MINOR(inode->i_rdev));
#endif

	if (!dp) {
		/* if not found, get out */
		return -EINVAL;
	}

	if (!dp->open) {
		/* if not yet opened, get out */
		return -EINVAL;
	}

	switch (ioctl_num) {
	case IXISA_GET_INFO:
		/* get device's information */
		ixisa_copy_devinfo((ixisa_devinfo_t *) ioctl_param, dp);
		break;
	case IXISA_SET_SIG:
		/* set signal for ixisa interrupt */
		if (set_signal((ixisa_signal_t *) ioctl_param, dp)) {
			return FAILURE;
		}
		break;
	case IXISA_READ_REG:
		/* read from register */
		if (read_reg((ixisa_reg_t *) ioctl_param, dp->base)) {
			return FAILURE;
		}
		break;
	case IXISA_WRITE_REG:
		/* write to register */
		if (write_reg((ixisa_reg_t *) ioctl_param, dp->base)) {
			return FAILURE;
		}
		break;
	default:
		return -EINVAL;
	}
	return SUCCESS;
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)
int ixa626_release(struct inode *inode, struct file *file)
#else
void ixa626_release(struct inode *inode, struct file *file)
#endif
{
	/* (export)
	 *
	 * This function is called by ixisa.o whenever a process attempts to
	 * closes the device file. It doesn't have a return value in kernel
	 * version 2.0.x because it can't fail (you must always be able to
	 * close a device).  In version 2.2.x it is allowed to fail.
	 *
	 * Arguments: read <linux/fs.h> for (*release) of struct file_operations
	 *
	 * Returned:  none */

	int minor;
	ixisa_devinfo_t *dp;

	minor = MINOR(inode->i_rdev);
	dp = align_to_minor(minor);

	if (!dp) {
		/* if not found, get out */
		return 0;
	}

	dp->open = 0;
	/* flag close */

	/* uninstall ISR */
	free_irq(dp->irq, dp);

	/* release io region */
	release_region(dp->base, IO_RANGE);

# if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0) && LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
        MOD_DEC_USE_COUNT;
#endif

# if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)
	return 0;
# endif
}

static struct file_operations fops = {
	open:ixa626_open,
	release:ixa626_release,
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,36)
        unlocked_ioctl:ixa626_ioctl,
#else
	ioctl:ixa626_ioctl,
#endif
};

int ixa626_open(struct inode *inode, struct file *file)
{
	/* (export)
	 *
	 * This function is called by ixisa.o whenever a process attempts to open
	 * the device file of A-626
	 *
	 * Arguments: read <linux/fs.h> for (*open) of struct file_operations
	 *
	 * Returned:  none */

	int minor;
	ixisa_devinfo_t *dp;
	ixisa_signal_t sig;

	minor = MINOR(inode->i_rdev);
	dp = align_to_minor(minor);

	if (!dp) {
		/* if not found, get out */
		return -EINVAL;
	}

	++(dp->open);
	if (dp->open > 1) {
		/* if already opened by someone, get out */
		--(dp->open);
		return -EBUSY;
	}

	/* clear signal setting */
	sig.sid = 0;
	sig.pid = 0;
	sig.task = 0;
	set_signal(&sig, dp);

	/* request io region */
	request_region(dp->base, IO_RANGE, MODULE_NAME);

	/* if irq is configured, install */
        if (dp->irq)
        {
                if (request_irq(dp->irq, irq_handler, SA_INTERRUPT | SA_SHIRQ, dp->name, dp))
                {
                        --(dp->open);
                        return -EBUSY;
                }
        }

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
        MOD_INC_USE_COUNT;
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,36)
        ixisa_minor[minor - 1] = minor;
        file->private_data = &ixisa_minor[minor - 1];
#endif
	return SUCCESS;
}

void cleanup_module()
{
	/* cleanup this module
	 *
	 * Arguments: none
	 *
	 * Returned:  none */

	ixisa_devinfo_t *dp;

	printk(KERN_INFO FAMILY ": %s ", MODULE_NAME);

	/* uninstall file operations */
	for (dp = dev; dp; dp = dp->next_f) {
		dp->fops = 0;
		off_flag(&dp->flags, IXISA_FLAG_ATTEND);
		printk(".");
	}
	printk(" has been removed.\n");
}

int init_module()
{
	/* initialize this module
	 *
	 * Arguments: none
	 *
	 * Returned:
	 *   integer 0 means ok, otherwise failed (module can't be load) */

	ixisa_devinfo_t *dp;

	printk(KERN_INFO FAMILY ": %s ", MODULE_NAME);

	if (!ixisa_dev) {
		printk("device list is empty!\n");
		return FAILURE;
	}

	/* align to first A-626 in ixisa list */
	for (dev = ixisa_dev; dev && dev->id != A626; dev = dev->next);

	if (dev->id != A626) {
		printk("fail!\n");
		return FAILURE;
	}

	/* install file operations */
	for (dp = dev; dp; dp = dp->next_f) {
                dp->fops = &fops;
		on_flag(&dp->flags, IXISA_FLAG_ATTEND);
		printk(".");
	}

	printk(" is ready.\n");

	return SUCCESS;
}
