/* Read ADC result with DMA transfer mode.

   Author: Reed Lai

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. */

/* File level history (record changes for this file here.)

   Set JP4 and JP6 to enable pacer and internal clock:

     JP4        JP6
     INTTRG     INTCLK
     +---+      +---+
     |+-+|      |+-+|
   1 ||o||    1 ||o||
     || ||      || ||
   2 ||o||    2 ||o||
     |+-+|      |+-+|
   3 | o |    3 | o |
     |   |      |   |
     +---+      +---+
     EXTTRG     EXTCLK

   Set JP7 and JP8 for DMA 1

     JP7             JP8
     DRQ             DACK
     +------+        +------+
     |+----+|        |+----+|
   1 ||o  o|| 2    1 ||o  o|| 2
     |+----+|        |+----+|
     | o  o |        | o  o |
     |      |        |      |
   5 | o  o | 6    5 | o  o | 6
     |      |        |      |
     +------+        +------+

   v 0.0.1  2 Dec 2002 by Reed Lai
     Improves signal operating.

   v 0.0.0 25 Feb 2002 by Reed Lai
     Create */

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include <signal.h>

#include "ixisa.h"
#include "_flag_inline.h"

#define AI_COUNT  10
#define WAIT_FOR_HARVEST 10

int fd;

enum {
	FLAG_HARVESTED,
	FLAG_FAILURE,
	FLAG_LAST
};
ixisa_flags_t flags = 0;

void sig_handler(int sig)
{
	ixisa_reg_t reg;
	unsigned int i;

	/* harvest ai dma data */
	if (ioctl(fd, IXISA_AIDMA_HARVEST)) {
		on_flag(&flags, FLAG_FAILURE);
		return;
	}

	/* flag when harvested */
	on_flag(&flags, FLAG_HARVESTED);
}

int main()
{
	char *dev_file;
	ixisa_reg_t reg;
	ixisa_signal_t sig;
	static struct sigaction act, act_old;
	ixisa_count_t count;
	ixisa_buf16_t ai;
	u_int16_t buf[AI_COUNT];
	unsigned int i;


	dev_file = "/dev/ixisa1";

	buf[4] = 0xffff;

	/* open device file */
	fd = open(dev_file, O_RDWR);
	if (fd < 0) {
		printf("Failure of open device file \"%s.\"\n", dev_file);
		return FAILURE;
	}

	/* signal action */
	act.sa_handler = sig_handler;
	sigemptyset(&act.sa_mask);
	sigaddset(&act.sa_mask, SIGALRM);
	if (sigaction(SIGALRM, &act, &act_old)) {
		close(fd);
		puts("Failure of signal action.");
		return FAILURE;
	}

	/* signal condiction */
	sig.sid = SIGALRM;
	sig.pid = getpid();
	if (ioctl(fd, IXISA_SET_SIG, &sig)) {
		close(fd);
		sigaction(SIGALRM, &act_old, NULL);
		puts("Failure of signal condiction.");
		return FAILURE;
	}

	/* gain for +/- 5V */
	reg.id = IXISA_ADGCR;
	reg.value = 0;
	if (ioctl(fd, IXISA_WRITE_REG, &reg)) {
		close(fd);
		sigaction(SIGALRM, &act_old, NULL);
		puts("Failure of gain.");
		return FAILURE;
	}

	/* channel 0 */
	reg.id = IXISA_ADMXCR;
	//reg.value = 0;
	if (ioctl(fd, IXISA_WRITE_REG, &reg)) {
		close(fd);
		sigaction(SIGALRM, &act_old, NULL);
		puts("Failure of channel.");
		return FAILURE;
	}
	usleep(23);
		/* delay 23us for channel switching */

	/* configure dma transfer mode */
	ai.buf = buf;
	ai.sz = AI_COUNT;
	ai.interval = 2000;  /* 2MHz / 2000 = 1 KHz */
	printf("ai counter = %d\n", ai.sz);
	if (ioctl(fd, IXISA_AIDMA_CONF, &ai)) {
		close(fd);
		sigaction(SIGALRM, &act_old, NULL);
		puts("Failure of dma configuration.");
		return FAILURE;
	}

	/* wait for dma harvest */
	while (!get_flag(&flags, FLAG_HARVESTED)) {
		if (get_flag(&flags, FLAG_FAILURE)) {
			close(fd);
			sigaction(SIGALRM, &act_old, NULL);
			puts("Failure of dma harvest.");
			return FAILURE;
		}
		usleep(WAIT_FOR_HARVEST);
	}

	/* dump */
	for (i = 0; i < AI_COUNT; i++)
		printf("%04x ", buf[i] & 0xfff);
	putchar('\n');

	close(fd);
	sigaction(SIGALRM, &act_old, NULL);
	puts("\nEnd of program.");
	return SUCCESS;
}
